# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scdlbot']

package_data = \
{'': ['*'], 'scdlbot': ['texts/*']}

install_requires = \
['bandcamp-downloader==0.0.13',
 'boltons==21.0.0',
 'ffmpeg-python==0.2.0',
 'mutagen>=1.46.0,<2.0.0',
 'patool==1.12',
 'plumbum>=1.8.0,<2.0.0',
 'prometheus_client>=0.15.0,<0.16.0',
 'python-telegram-bot==12.8',
 'python-telegram-handler==2.2.1',
 'requests[socks]>=2.28.1,<3.0.0',
 'scdl==2.7.3',
 'yt-dlp==2022.11.11']

entry_points = \
{'console_scripts': ['scdlbot = scdlbot.__main__:main']}

setup_kwargs = {
    'name': 'scdlbot',
    'version': '0.14.1',
    'description': 'Telegram Bot for downloading MP3 rips of tracks/sets from SoundCloud, Bandcamp, YouTube with tags and artwork',
    'long_description': 'Music Downloader Telegram Bot aka scdlbot\n=========================================\n\n\n.. image:: https://img.shields.io/badge/telegram-@scdlbot-blue.svg\n        :target: https://t.me/scdlbot\n        :alt: Telegram Bot\n\n.. image:: https://img.shields.io/github/license/gpchelkin/scdlbot.svg\n        :target: https://github.com/gpchelkin/scdlbot/blob/master/LICENSE\n        :alt: MIT License\n\n.. image:: https://readthedocs.org/projects/scdlbot/badge/?version=latest\n        :target: https://scdlbot.readthedocs.io/\n        :alt: Documentation Status\n\n.. image:: https://img.shields.io/pypi/v/scdlbot.svg\n        :target: https://pypi.org/project/scdlbot\n        :alt: PyPI Version\n\n.. image:: https://github.com/gpchelkin/scdlbot/workflows/build/badge.svg?branch=master&event=push\n        :target: https://github.com/gpchelkin/scdlbot/actions/workflows/build.yml\n        :alt: GitHub Actions Build Status\n\n.. image:: https://github.com/gpchelkin/scdlbot/actions/workflows/codeql-analysis.yml/badge.svg?branch=master&event=push\n        :target: https://github.com/gpchelkin/scdlbot/actions/workflows/codeql-analysis.yml\n        :alt: GitHub Actions CodeQL Status\n\n.. image:: https://deepsource.io/gh/gpchelkin/scdlbot.svg/?label=active+issues&show_trend=true\n        :target: https://deepsource.io/gh/gpchelkin/scdlbot/?ref=repository-badge\n        :alt: DeepSource Active Issues\n\n.. image:: https://deepsource.io/gh/gpchelkin/scdlbot.svg/?label=resolved+issues&show_trend=true\n        :target: https://deepsource.io/gh/gpchelkin/scdlbot/?ref=repository-badge\n        :alt: DeepSource Resolved Issues\n\n.. image:: https://codeclimate.com/github/gpchelkin/scdlbot/badges/issue_count.svg\n        :target: https://codeclimate.com/github/gpchelkin/scdlbot\n        :alt: Code Climate Issue Count\n\n.. image:: https://api.codacy.com/project/badge/Grade/7dfb6d8e7a094987b303e9283fc7368c\n        :target: https://app.codacy.com/gh/gpchelkin/scdlbot\n        :alt: Codacy Build Status\n\n.. image:: https://codebeat.co/badges/57243b9d-2269-4f31-a35b-6aedd11626d2\n        :target: https://codebeat.co/projects/github-com-gpchelkin-scdlbot-master\n        :alt: Codebeat Quality\n\n.. image:: https://www.codefactor.io/repository/github/gpchelkin/scdlbot/badge\n        :target: https://www.codefactor.io/repository/github/gpchelkin/scdlbot\n        :alt: CodeFactor\n\nTelegram Bot for downloading MP3 rips of tracks/sets from\nSoundCloud, Bandcamp, YouTube with tags and artwork.\n\n\n* Free software: `MIT License <https://github.com/gpchelkin/scdlbot/blob/master/LICENSE>`__\n* Documentation: https://scdlbot.readthedocs.io\n\n\n.. contents:: :depth: 2\n\n\nscdlbot Usage in Telegram\n-------------------------\n\nSend ``/start`` or ``/help`` command to bot\nor refer directly to the `help message <scdlbot/texts/help.tg.md>`__.\n\nPlease report all bugs and issues and suggest your improvements\nto `issues <https://github.com/gpchelkin/scdlbot/issues>`__.\n\nSupported sites and mainly used packages\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nscdlbot is standing on the shoulders of giants:\n\n-  `Telegram Bot API <https://core.telegram.org/bots/api>`__:\n   `python-telegram-bot <https://github.com/python-telegram-bot/python-telegram-bot>`__\n-  `SoundCloud <https://soundcloud.com>`__:\n   `scdl <https://github.com/flyingrub/scdl>`__\n-  `Bandcamp <https://bandcamp.com>`__:\n   `bandcamp-dl <https://github.com/iheanyi/bandcamp-dl>`__\n-  `YouTube <https://www.youtube.com/>`__, `Yandex.Music <https://music.yandex.com/>`__,\n   `Mixcloud <https://www.mixcloud.com/>`__, and almost everything from this `list <https://yt-dl.org/supportedsites.html>`__:\n   `youtube-dl <https://yt-dl.org>`__\n\nRun your own scdlbot\n--------------------\n\nInstallation & Configuration\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nRequirements\n""""""""""""\n\nThose should be available in your ``PATH``:\n\n-  `Python 3.9+ <https://www.python.org>`__\n   (`pyenv <https://github.com/pyenv/pyenv>`__ and `poetry <https://python-poetry.org/>`__ are recommended)\n-  `FFmpeg 4.0+ <https://ffmpeg.org/download.html>`__ if not running on Heroku\n   (fresh builds for\n   `Windows <https://www.gyan.dev/ffmpeg/builds/#release-builds>`__,\n   `macOS <https://evermeet.cx/ffmpeg/>`__\n   and `Linux <https://johnvansickle.com/ffmpeg/>`__ are recommended)\n-  `Heroku CLI <https://cli.heroku.com>`__ is recommended if you want to deploy to Heroku\n\nInstall / Update stable from `PyPI <https://pypi.org/project/scdlbot>`__ (recommended)\n""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""\n\n::\n\n    pip3 install scdlbot --upgrade\n\n...or get unstable from `Git source repository <https://github.com/gpchelkin/scdlbot>`__\n""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""\n\n::\n\n    git clone https://github.com/gpchelkin/scdlbot.git\n    cd scdlbot\n    pip3 install --requirement requirements.txt\n\n    # Update:\n    git pull\n    pip3 install --requirement requirements.txt\n\n\nConfigure Bot\n"""""""""""""\n\nDownload or copy configuration file sample and set up\nconfig environment variables in it:\n\n::\n\n    # If you\'ve installed from PyPI - download sample config somewhere:\n    curl -o .env https://raw.githubusercontent.com/gpchelkin/scdlbot/master/.env.sample\n\n    # If you\'ve got Git source - just copy sample config:\n    cp .env.sample .env\n\n    # Use your favourite editor. Sample config is self-documented:\n    nano .env\n\nTelegram Bot Settings\n"""""""""""""""""""""\n\nSend the commands from respective filenames in ``telegram_settings`` dir to `@BotFather <https://t.me/BotFather>`__, choose your bot and copy corresponding values in order to use the bot conveniently.\nDisable privacy mode if you want bot to read and check every message in group for links.\nOtherwise, it would work only for commands.\n\nRunning Locally or at Dedicated Server\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nUsing `Heroku Local <https://devcenter.heroku.com/articles/heroku-local#run-your-app-locally-using-the-heroku-local-command-line-tool>`__ (preferred)\n"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""\n\nYou will need `Heroku CLI <https://cli.heroku.com/>`__ installed.\n\n::\n\n    # If you\'ve installed from PyPI - download Procfile first (otherwise already present in Git repository):\n    curl -O https://raw.githubusercontent.com/gpchelkin/scdlbot/master/Procfile\n\n    # For long polling mode (when USE_WEBHOOK=0):\n    heroku local -e .env worker\n    # For webhook mode (when USE_WEBHOOK=1):\n    heroku local -e .env web\n\nUsing Python only\n"""""""""""""""""\n\n::\n\n    export $(grep -v \'^#\' .env | xargs)\n    python3 -m scdlbot\n    # or in one line:\n    env $(grep -v \'^#\' .env | xargs) python3 -m scdlbot\n\n    # If you\'ve installed package from PyPI into the system,\n    # you can also replace \'python3 -m scdlbot\' with only \'scdlbot\'\n\nDeploying to `Heroku <https://heroku.com/>`__\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n|Deploy|\n\nRegister on Heroku, press the button above and\nconfigure variables for deploying.\nWhen app is deployed you **must** set only one dyno working on\n"Resources" tab in your app settings depending on `which way of getting\nupdates <https://core.telegram.org/bots/api#getting-updates>`__ you have\nchosen and set in config variables: ``worker`` for polling or ``web``\nfor webhook.\n\nManually\n""""""""\n\nYou can do the same as the button above but using `Heroku\nCLI <https://cli.heroku.com/>`__, not as much of a fun. Assuming you are in\n``scdlbot`` repository directory:\n\n::\n\n    heroku login\n    # Create app with Python 3 buildpack and set it for upcoming builds:\n    heroku create --buildpack heroku/python\n    heroku buildpacks:set heroku/python\n    # Add FFmpeg buildpack needed for youtube-dl & scdl:\n    heroku buildpacks:add --index 1 https://github.com/jonathanong/heroku-buildpack-ffmpeg-latest.git --app scdlbot\n    # Deploy app to Heroku:\n    git push heroku master\n    # Set config vars automatically from your local .env file\n    heroku plugins:install heroku-config\n    heroku config:push\n    # Or set them manually:\n    heroku config:set TG_BOT_TOKEN="<TG_BOT_TOKEN>" STORE_CHAT_ID="<STORE_CHAT_ID>" ...\n\nIf you use webhook, start web dyno and stop worker dyno:\n\n::\n\n    heroku ps:scale web=1 worker=0\n    heroku ps:stop worker\n\nIf you use polling, start worker dyno and stop web dyno:\n\n::\n\n    heroku ps:scale worker=1 web=0\n    heroku ps:stop web\n\nSome useful commands:\n\n::\n\n    # Attach to logs:\n    heroku logs -t\n    # Test run ffprobe\n    heroku run "ffprobe -version"\n\nDeploying to `Dokku <https://github.com/dokku/dokku>`__\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nUse Dokku (your own Heroku) installed on your own server.\nApp is tested and fully ready for deployment with polling\n(no webhook yet).\nhttps://github.com/dokku/dokku-letsencrypt\n\n::\n\n    export DOKKU=<your_dokku_server>\n    scp .env $DOKKU:~\n    ssh $DOKKU\n        export DOKKU=<your_dokku_server>\n        dokku apps:create scdlbot\n        dokku certs:generate scdlbot scdlbot.$DOKKU\n        dokku config:set scdlbot $(cat .env | xargs)\n        logout\n    git remote add dokku dokku@$DOKKU:scdlbot\n    git push dokku master\n    ssh $DOKKU\n        dokku ps:scale scdlbot worker=1 web=0\n        dokku ps:restart scdlbot\n\n.. |Deploy| image:: https://www.herokucdn.com/deploy/button.svg\n    :target: https://heroku.com/deploy\n',
    'author': 'George Pchelkin',
    'author_email': 'george@pchelk.in',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/gpchelkin/scdlbot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
